/*
*****************************************
* jvmData - JVM memory access functions *
*****************************************
*/
#include <propeller.h>
#include "jvmEeprom.h"
#include "jvmDefines.h"
#include "jvmData.h"

  
//global variables
int  javaProg;                          //address of javaProg storage area
int  javaSize;                          //size of javaProg
  
int  jvmHP;                             //heap pointer
int  jvmSP;                             //stack pointer
int  jvmSB;                             //stack base (also first free eeprom address)
int  jvmFP;                             //frame pointer
int  jvmPC;                             //program counter
int  jvmMP;                             //method pointer
int  jvmEP;                             //exception pc

int  memSize;                           //requested by QueryConfig
int  progSize;                          //requested by QueryConfig
int  heapBase;                          //requested by QueryConfig
int  staticBase;                        //base of static variables area
int  stringIndex;                       //address of string index table
int  eeBase;                            //requested by stamp\core\EEPROM.java
                                        //  set to return true available free eeprom space
                                       
int  bpList[BREAKPOINTS];               //list of breakpoints (entries are jvmPC address values)
char  numBreakpoints;                   //number of breakpoints
char  activity;                         //requested by QueryStatus
char  carry;                            //set to true if math expression overflows
char  kComEndLoad;                      //holds device version (0x50,0x70,0x78)
char  notifyStepDone;                   //set to true if jvm must notify IDE after step is done 
char  mustDoStep;                       //set to true when jvm must do step
char  eeByte;                           //to support eeprom read/write
  
#define false FALSE
#define true TRUE
  
long debugger(void) {
/* return true if javaprogram will be debugged */
  return *(char *)javaProg == kDebugMagic;
}
  
void writeJavaByte(long address, long value) {
/* write value to java program */
  *(char *)(javaProg + address) = value;
}
  
long readJavaByte(long address) {
/* read value from java program */
  return *(char *)(javaProg + address) & 0xFF;
}

void writeJavaIntBE(long address, long value) {
/* write value to java program using big endian */
  *(char *)(javaProg + address) = value>>8;
  *(char *)(javaProg + address + 1) = value;
}
  
long readJavaIntBE(long address) {
  long value;
/* read value from java program using big endian */
  value = *(char *)(javaProg + address);
  value <<= 8;
  value += *(char *)(javaProg + address + 1);
  return value & 0xFFFF;
}  

void init(long jprog, long jsize, long port) {
  long i;
/* initialize jvm registers */
  javaProg = jprog;
  javaSize = jsize;
  switch (port) {
    case 1: kComEndLoad = 0x78; //download via propeller TX/RX pins (no loopback, no echo)
            break;
	case 2: kComEndLoad = 0x70; //download via spin stamp SOUT/SIN pins (loopback and echo)
            break;
	//default:
  }
  //set some constant fields
  memSize = javaSize;               //total space for java program
  progSize = 0x2000;                //constant value, does not reflect true program size
  heapBase = javaSize-1;            //highest heap address
  eeBase = (heapBase - javaSize);   //Set eeBase to reflect true free eeprom space
 
  //clear variables
  for (i=0; i<BREAKPOINTS; i++) {
    bpList[i] = 0;
  }
  numBreakpoints = 0;
  if (debugger()) activity = kPassive;
  else activity = kRun;
  carry = false;
  notifyStepDone = false;
  jvmEP = 0;
      
  //initialize jvm registers
  jvmMP = readJavaIntBE(JVM_MAIN);                    //first method is 'main'
  jvmPC = readJavaIntBE(jvmMP + JVM_CODE);            //Get address for 'main' code
  jvmSB = readJavaIntBE(JVM_STKBASE);                 //SB loaded with stackbase

  //clear heap/stack area
  for (i=jvmSB; i<memSize; i++) {
    writeJavaByte(i,0);
  }
    
  jvmFP = jvmSB;                                      //FP starts at initial stackframe
  jvmSP = jvmSB;                                      //SP starts at initial stackframe
  jvmSP += (readJavaByte(jvmMP + JVM_PARAMETERS) & 0x7F);      //skip parameters
  jvmSP += readJavaByte(jvmMP + JVM_LOCALS);          //skip locals
  
  writeJavaIntBE(jvmSP+0,0xFFFF);                     //old mp
  writeJavaIntBE(jvmSP+2,0x0000);                     //old fp
  writeJavaIntBE(jvmSP+4,0x0006);                     //old pc (= Object method ???)
  jvmSP += 6;                                         //skip saved pointers

  jvmHP = readJavaByte(JVM_STATIC);                   //statics allocated from heapBase
  if (jvmHP == 0)
    jvmHP += 256;
  jvmHP = heapBase - jvmHP;                           //HP starts just before statics
  staticBase = jvmHP;                                 //mark start of static variables are
  stringIndex = readJavaIntBE(JVM_STRTABLE);          //mark start of string index table
}  

long getLdc(long index) {
/* read value from String Index Table */
  return readJavaIntBE(stringIndex + (index<<1));
}
  
long getStatic(long offset) {
/* read static int variable */
  return readJavaIntBE(staticBase + offset);
}

void putStatic(long offset, long value) {
/* write static int variable */
  writeJavaIntBE(staticBase + offset,value);
}

void eeBlockWrite(long address) {
/* write block of 16 bytes from javaProg to eeprom */
  VarBackup(javaProg+address,javaProg+address+15);
}

void eeWriteByte(long address, long value) {
/* write byte to javaProg eeprom */
  eeByte = value;
  FromRam((long)&eeByte,(long)&eeByte,javaProg+address);
}

long eeReadByte(long address) {
/* read byte from javaProg eeprom */
  ToRam((long)&eeByte,(long)&eeByte,javaProg+address);
  return eeByte & 0xFF;
}  

long getEndLoad(void) {
/* get value for kComEndLoad */
  return kComEndLoad & 0xFF;
}
  
long getMemSize(void) {
/* get memory size */
  return memSize & 0xFFFF;
}

long getProgSize(void) {
/* get program size */
  return progSize & 0xFFFF;
}

long getHeapBase(void) {
/* get heap base */
  return heapBase & 0xFFFF;
}
    
void setMustDoStep(void) {
/* set must do step */
  mustDoStep = 0xFF;
}

void clearMustDoStep(void) {
/* clear must do step */
  mustDoStep = 0x00;
}

long mustStep(void) {
/* test if must do step */
  return mustDoStep == 0xFF;
}


void setNotifyStepDone(void) {
/* set step */
  notifyStepDone = 0xFF;
}

void clearNotifyStepDone(void) {
/* clear step */
  notifyStepDone = 0x00;
}

long mustNotifyStepDone(void) {
/* test notifyStepDone */
  return notifyStepDone == 0xFF;
}

void setCarry(void) {
/* set carry */
  carry = 0xFF;
}

void clearCarry(void) {
/* clear carry */
  carry = 0x00;
}

long isCarry(void) {
/* test carry */
  return carry == 0xFF;
}

void setJvmEP(long ep) {
/* set exception pointer */
  jvmEP = ep;
}

long getJvmEP(void) {
/* get exception pointer */
  return jvmEP & 0xFFFF;
}

void setJvmHP(long hp) {
/* set heap pointer */
  jvmHP = hp;
}

long getJvmHP(void) {
/* get heap pointer */
  return jvmHP & 0xFFFF;
}

void setJvmSP(long sp) {
/* set stack pointer */
  jvmSP = sp;
}

long getJvmSP(void) {
/* get stack pointer */
  return jvmSP & 0xFFFF;
}

void setJvmSB(long sb) {
/* set stack base */
  jvmSB = sb;
}

long getJvmSB(void) {
/* get stack base */
  return jvmSB & 0xFFFF;
}

void setJvmFP(long fp) {
/* set frame pointer */
  jvmFP = fp;
}

long getJvmFP(void) {
/* get frame pointer */
  return jvmFP & 0xFFFF;
}

void setJvmPC(long pc) {
/* set program counter */
  jvmPC = pc;
}

long getJvmPC(void) {
/* get program counter */
  return jvmPC & 0xFFFF;
}

void incJvmPC(long value) {
/* add value to jvmPC */
  jvmPC += value;
}
   
void setJvmMP(long mp) {
/* set method pointer */
  jvmMP = mp;
}

long getJvmMP(void) {
/* get method pointer */
  return jvmMP & 0xFFFF;
}

void setActivity(long kActivity) {
/* set jvm state to kPassive,kRun,kJog,kBreak */
  activity = kActivity;
}

long getActivity(void) {
/* get current activity state */
  return activity & 0xFF;
}

void setBreakpointByte(long index, long value) {
/* write byte to bpList */
  *(char *)((long)bpList + index) = value;
}
  
void setBreakpoints(long value) {
/* set number of breakpoints */
  numBreakpoints = value;
}

long isBreakpoint(void) {
  long i,n;
/* check if pc is a breakpoint */
  n = numBreakpoints;
  i = 0;
  while (n > 0) {
    if ((bpList[i] >= jvmPC-3) && (bpList[i] < jvmPC)) //breakpoints are start addresses of a jemcode instruction
      return true;                                     //and we want to stop AFTER the instruction
    i++;
	n--;
  }
  return false;
}

  
void writeJVMregister(long address, long value) {
  long mask;
/* write value to register */
  address &= 0xFF;
  switch (address) {
    case nmPortA: mask = OUTA & 0xFFFFFF00;
	              mask |= value;
	              OUTA = mask;
	              break;
    case nmPortB: mask = OUTA & 0xFFFF00FF;
	              mask |= value<<8;
	              OUTA = mask;
	              break;
    case nmPortC: mask = OUTA & 0xFF00FFFF;
	              mask |= value<<16;
	              OUTA = mask;
	              break;
    case nmPortD: mask = OUTA & 0x00FFFFFF;
	              mask |= value<<24;
	              OUTA = mask;
	              break;
    //default:
  }
}
  
long readJVMregister(long address) {
  long value;
/* read value from register */
  value = 0;
  address &= 0xFF;
  switch (address) {
    case nmPortA: value = INA;
	              break;
    case nmPortB: value = INA >> 8;
	              break;
    case nmPortC: value = INA >> 16;
	              break;
    case nmPortD: value = INA >> 24;
	              break;
    case JVM_SB_MSB: value = ((0x8000 - javaSize) + jvmSB) >> 8;  // Highbyte of simulated stackbase (first free eeprom address)
                     break;
	case JVM_SB_LSB: value = ((0x8000 - javaSize) + jvmSB);       // Lowbyte of simulated stackbase (first free eeprom address)
                     break;
	case JVM_HEAP_POINTER_MSB: value = jvmHP >> 8;  //jvm heap pointer (highbyte)
                               break;
	case JVM_HEAP_POINTER_LSB: value = jvmHP;       //jvm heap pointer (lowbyte)
                               break;
	case JVM_STACK_POINTER_MSB: value = jvmSP >> 8; //jvm stack pointer (highbyte)
                                break;
	case JVM_STACK_POINTER_LSB: value = jvmSP;      //jvm stack pointer (lowbyte)
                                break;
	//default:
  }
  return value & 0xFF;
}  

   