/*
*********************************
* File jvmEeprom: access eeprom *
*********************************
*/

#include <propeller.h>
#include "jvmEeprom.h"

static void SendAck(long ackbit) {
  // Transmit an acknowledgement bit (ackbit).
  OUTA |= (ackbit & 1) << SDA; // Set SDA output state to ackbit
  DIRA |= 1 << SDA;            // Make sure SDA is an output
  OUTA |= 1 << SCL;            // Send a pulse on SCL
  OUTA &= ~(1<<SCL);           //
  DIRA &= ~(1<<SDA);           // Let go of SDA
}

static long GetAck(void) {
  long ackbit;
  // GetByte and return acknowledge bit transmitted by EEPROM after it receives a byte.
  // 0 = ACK, 1 = NACK.
  DIRA &= ~(1<<SDA);         // SDA input so 24LC256 controls
  OUTA |= 1<<SCL;            // Start a pulse on SCL
  ackbit = (INA >> SDA) & 1; // Get the SDA state from 24LC256
  OUTA &= ~(1<<SCL);         // Finish SCL pulse
  OUTA &= ~(1<<SDA);         // SDA will hold low
  DIRA |= 1<<SDA;            // SDA output, master controls
  return ackbit;
}
  
static void I2cStart(void) {
  // I2C start condition.  SDA transitions from high to low while the clock is high.
  // SCL does not have the pullup resistor called for in the I2C protocol, so it has to be
  // set high. (It can't just be set to inSendByte because the resistor won't pull it up.)
  OUTA |= 1<<SCL;    // SCL pin output
  DIRA |= 1<SCL;     //
  DIRA &= ~(1<<SDA); // Let pulled up SDA pin go high
  OUTA &= ~(1<<SDA); // Transition SDA pin low
  DIRA |= 1<<SDA;    // SDA output
}

static void I2cStop(void) {
  // Send I2C stop condition.  SCL must be high as SDA transitions from low to high.
  // See note in i2cStart about SCL line.
  OUTA &= ~(1<<SDA); // SDA low
  DIRA |= 1<<SDA;    // SDA output
  OUTA |= 1<<SCL;    // SCL high
  DIRA &= ~(1<<SDA); // SDA input
}
   
static long GetByte(void) {
  long value,i;
  // Shift in a byte msb first.  
  value = 0; // Clear value
  DIRA &= ~(1<<SDA);    // SDA input so 24LC256 can control
  for (i=0; i<8; i++) { // Repeat shift in eight times
    OUTA |= 1<<SCL;     // Start an SCL pulse
    value <<= 1;        // Shift the value left
    value += (INA >> SDA) & 1; // Add the next most significant bit
    OUTA &= ~(1<<SCL); // Finish the SCL pulse
  }
  return value;
}

static long SendByte(long b) {
  long i;
  // Shift a byte to EEPROM msb first.  Return if EEPROM acknowledged.  Returns
  // acknowledge bit.  0 = ACK, 1 = NACK.
  OUTA &= ~(1<<SCL);        // SCL low, SDA can change
  for (i=0; i<8; i++) {     // 8 reps sends 8 bits
    if (b & 0x80) OUTA |= 1<<SDA;
	else OUTA &= ~(1<<SDA); // Lowest bit sets state of SDA
    OUTA |= 1<<SCL;         // Pulse the SCL line
    OUTA &= ~(1<<SCL);      //
    b <<= 1;                // Shift b left for next bit
  }
  return GetAck();          // Call GetByteAck and return EEPROM's Ack
}

static void Poll(void) {
  long ackbit;
  // Poll until acknowledge.  This is especially important if the 24LC256 is copying from
  // buffer to EEPROM.
  ackbit = 1;                // Make acknowledge 1
  do {                       // Send/check acknowledge loop
    I2cStart();              // Send I2C start condition
    ackbit = SendByte(0xA0); // Write command with EEPROM's address
  } while (ackbit);          // Repeat while acknowledge is not 0
}
  
static void SetAddr(long addr) {
  //Sets EEPROM internal address pointer.
  Poll();              // Poll until EEPROM acknowledges
  SendByte(addr >> 8); // Send address high byte
  SendByte(addr);      // Send address low byte
}
  
void FromRam(long startAddr, long endAddr, long eeStart) {
  long addr, page, eeAddr;
  //Copy startAddr..endAddr from main RAM to EEPROM beginning at eeStart address.
  addr = startAddr;                     // Initialize main RAM index
  eeAddr = eeStart;                     // Initialize EEPROM index
  do {
    page = addr+64-eeAddr % 64;
	if (page > endAddr+1) page = endAddr+1; // Find next EEPROM page boundary
    SetAddr(eeAddr);                    // Give EEPROM starting address
    do {                                // Bytes -> EEPROM until page boundary
      SendByte(addr++);
    } while ((addr != page) && (addr <= endAddr));
    I2cStop();                          // From 24LC256's page buffer -> EEPROM
    eeAddr = addr-startAddr+eeStart;    // Next EEPROM starting address
  } while (addr <= endAddr);            // Quit when RAM index > end address
} 
  
void ToRam(long startAddr, long endAddr, long eeStart) {
  char *addr;
  //Copy from EEPROM beginning at eeStart address to startAddr..endAddr in main RAM.
  SetAddr(eeStart);                     // Set EEPROM's address pointer 
  I2cStart(); 
  SendByte(0xA1);                       // EEPROM I2C address + read operation
  if (startAddr == endAddr)
    addr = (char *)startAddr;
  else
    for (addr = (char *)startAddr; addr<(char *)endAddr; addr++) { // Main RAM index startAddr to endAddr
      *addr = GetByte();                // Get byte from EEPROM & copy to RAM 
      SendAck(ACK); 
	}                                   // Acknowledge byte received
  *addr = GetByte();                    // Get byte from EEPROM & copy to RAM 
  SendAck(NACK); 
  I2cStop();                            // Stop sequential read
}
  
void VarBackup(long startAddr, long endAddr) {
  long addr, page, eeAddr;
  //Copy contents of address range defined by startAddr..endAddr from main RAM to EEPROM.
  FromRam(startAddr, endAddr, startAddr); // Pass addresses to the Write method
}
  
void VarRestore(long startAddr, long endAddr) {
  long addr;
  //Copy contents of address range defined by startAddr..endAddr from EEPROM to RAM.
  ToRam(startAddr, endAddr, startAddr); // Pass addresses to the read method
}



