/**
 * This library provides formatted string output methods based on the
 * standard C sprintf function.  With this class you can convert
 * byte, word and long types into binary, octal, decimal or hexadecimal formatted
 * strings.  You can specify the minimum and maximum number of columns to
 * display your values, and these values can be left or right justified, with
 * or without padded zeros.  Methods are included for the following C library
 * functions: itoa, atoi and sprintf.
 *
 * This implementation defines the function bprintf that take one,
 * and only one, variable parameter. The original format string must be split
 * into pieces that all have one format specifier. This normally is not a
 * problem since most format strings consists of fixed text with format
 * specifiers for values.
 *
 * @version 1.2 Dec 16, 2007
 * @author Peter Verkaik (verkaik6@zonnet.nl)
 */

#include <string.h>
#include "jvmDefines.h"
#include "jvmFormat.h"

char buf[33]; //for integer conversions (base2 is 32 binary digits plus closing null)

  
long isdigit(long ch) {
  /**
   * Test if character is a digit.
   *
   * @param ch Character to be tested
   * @return True if ch is a digit
   */
  return (ch <= '9') && (ch >= '0');
}

long isspace(long ch) {
  /**
   * Test if character is a space (0x20), tab (\t) or newline (\n).
   *
   * @param ch Character to be tested
   * @return True if ch is a space, tab or newline
   */
  return (ch == 32) || (ch == 9) || (ch == 10);
}

void reverse(long s) {
  long c,k;
  /**
   * Reverse string in place.
   *
   * @param s Character array to reverse
   */
  k = s + strlen((char *)s) - 1; //address of last character
  while (s < k) {
    c = *(char *)s;
    *(char *)s++ = *(char *)k;
    *(char *)k-- = c;
  }
}

void itoab(long n, long s, long b) {
  long lowbit,sorg;
  /*
   * Convert unsigned integer to unsigned string using specific base.
   * (This is a non-standard function). Used by <code>printf</code>.
   *
   * @param n Integer value to convert
   * @param s Character array to hold output
   * @param b Base for output (2=binary, 8=octal, 10=decimal, 16=hexadecimal)
   */
  sorg = s;
  b = (b >> 1) & 0x0F;
  do {                             // generate digits/letters in reverse order
    lowbit = n & 1;
    n = (n >> 1) & 0x7FFFFFFF;
    *(char *)s = ((n % b) << 1) + lowbit;
    if (*(char *)s < 10)
      *(char *)s += '0';
    else
      *(char *)s += ('A'-10);
    ++s;
    n = n/b;
  } while (n != 0);
  *(char *)s = 0;                        // closing 0
  reverse(sorg);
}

void itoa(long n, long s) {
  long sign,sorg;
  /**
   * Convert signed integer to signed decimal string.
   *
   * @param n Integer value to convert
   * @param s Character array to hold output
   */
  sorg = s;
  sign = n;                              // record sign
  if (n < 0)                              
    n = -n;                              // make n positive
  do {                                   // generate digits in reverse order
    *(char *)s++ = ((n%10) + '0');       // get next digit
    n = n/10;
  } while (n != 0);
  if (sign < 0)
    *(char *)s++ = '-';                  // sign character only if negative
  *(char *)s = 0;                        // closing 0
  reverse(sorg);
}

long atoi(long s) {
  long val,sign,n;
  /**
   * Convert signed decimal string to signed integer.
   *
   * @param s Character array that holds decimal string
   * @return Value of signed decimal string
   */
  while (isspace(*(char *)s))
    ++s;                                 // skip over any white space
  sign = 1;                              // assume positive
  switch (*(char *)s) {
    case '-': sign = -1;                 // is negative
              ++s;                       // advance only if sign (- or +) present
              break;
	case '+': ++s;
	          break;
	//default:
  }
  n = 0;
  while (isdigit(*(char *)s))
    n = (10*n) + (*(char *)s++) - '0';  // calculate value
  val = sign * n;                        // adjust for sign
  return val;
}
 
long bprintf(long str, long si, long fmt, long arg) {
  long index,pad,len,maxlen,minlen,bi,left,strtype,sorg;
  /**
   * Print formatted string to character buffer, maintaining buffer index.
   * (special version of sprintf, to ease porting printf and sprintf with multiple arguments)
   *
   * format specifiers for byte, word, long: %c,%d,%i,%b,%o,%u,%x
   * format specifier for string: %s
   * field specification (example)
   *   %-05.7d where - for left justify, default right justify
   *                 0 for padding with zeroes, default spaces
   *                 5 minimum field width
   *                 . field separator
   *                 7 maximum field width
   *
   * An original C sprintf statement like:
   * sprintf(buffer,"outside temperature %d %s inside temperature %d %s",12,"fahrenheit",24,"celsius");
   * would be written as:
   * OBJ
   *   fmt: "Format"
   * VAR
   *   byte buffer[128]
   *   word k
   * PUB
   *   k := fmt.bprintf(@buffer,0,string("outside temperature %d "),12)
   *   k := fmt.bprintf(@buffer,k,string("%s "),string("fahrenheit"))
   *   k := fmt.bprintf(@buffer,k,string("inside temperature %d "),24)
   *   k := fmt.bprintf(@buffer,k,string("%s"),string("celsius"))
   *   buffer[k] := 0
   * The original sprintf is simply split up in smaller bprintf statements that take only 1 argument.
   * The end result is the same: a single output string in buffer.
   *
   * @param str Character array for formatted output
   * @param si Start index in str
   * @param fmt String defining format
   * @param arg value parameter
   * @return Index in str pointing beyond last written character
   */
  sorg = str; //save start address of output buffer
  len = 0;
  str += si; //start from index
  while (*(char *)fmt != 0) {
    if (*(char *)fmt != '%') {
      if (*(char *)fmt != '\\') {
        *(char *)str++ = *(char *)fmt++;
      }
	  else {
        fmt++; //skip backslash
        switch (*(char *)fmt) {
          case '\\':*(char *)str++ = '\\';
                    fmt++;
					break;
          case 't': *(char *)str++ = 0x09;
                    fmt++;
					break;
          case 'n': *(char *)str++ = 0x0A;
                    fmt++;
					break;
          case 'r': *(char *)str++ = 0x0D;
		            fmt++;
					break;
          default: *(char *)str++ = '\\'; //output skipped backslash
                   *(char *)str++ = *(char *)fmt++;
        }
	  }
	  continue;
    }
	else {
      fmt++; //skip control character %
    }
	if (*(char *)fmt == '%') {
      *(char *)str++ = *(char *)fmt++; //%%
      continue;
    }
	if (*(char *)fmt == '-') {
      left = TRUE;
      ++fmt; //skip -
    }
	else {
      left = FALSE;
    }
	if (*(char *)fmt == '0') {
      pad = '0';
    }
	else {
      pad = ' ';
    }
	if (isdigit(*(char *)fmt)) {  // minimum field width
      bi = 0;
      while (isdigit(*(char *)fmt))
        buf[bi++] = *(char *)fmt++;
      buf[bi] = 0;
      minlen = atoi((long)buf);
    }
	else {
      minlen = 0;
    }
	if (*(char *)fmt == '.') {   // maximum field width
      ++fmt; //skip .
      bi = 0;
      while (isdigit(*(char *)fmt))
        buf[bi++] = *(char *)fmt++;
      buf[bi] = 0;
      maxlen = atoi((long)buf);
    }
	else {
      maxlen = 0;
    }
	strtype = FALSE; //assume no string value
    switch (*(char *)fmt++) {
      case 'c': buf[0] = arg;       // character
                buf[1] = 0;
                len = 1;
				break;
      case 'd':
	  case 'i': itoa(arg, (long)buf);  // signed decimal
                len = strlen(buf);
				break;
      case 'b': itoab(arg, (long)buf, 2);  // binary
                len = strlen(buf);
				break;
      case 'o': itoab(arg, (long)buf, 8);  // octal
                len = strlen(buf);
				break;
      case 'u': itoab(arg, (long)buf, 10); // unsigned decimal
                len = strlen(buf);
				break;
      case 'x': itoab(arg, (long)buf, 16); // hexadecimal
                len = strlen(buf);
				break;
      case 's': strtype = TRUE;
                len = strlen((char *)arg); // string
				break;
      default:  buf[0] = 0;       // no valid format specifier
                len = 0;
    }
	if ((maxlen != 0) && (maxlen < len))
      len = maxlen;
    if (minlen > len)
      minlen = minlen - len;
    else
      minlen = 0;
    bi = 0;
    if (left == FALSE) {
      if ((buf[bi] == '-') && (pad == '0')) {
        *(char *)str++ = buf[bi++];
        len--;
      }
	  while (minlen > 0) {
        minlen--;
        *(char *)str++ = pad;
      }
	}
	while (len > 0) {
      len--;
      if (strtype == FALSE)
        *(char *)str++ = buf[bi++];  //copy ascii string of value
      else {
        if (*(char *)arg == 0)
          break;
        *(char *)str++ = *(char *)arg++;  //copy string argument
      }
	}
	if (left == TRUE)
      while (minlen > 0) {
        minlen--;
        *(char *)str++ = pad;
      }
  }
  return str-sorg;
}

long sprintf(long str, long fmt, long arg) {
  long index;
  /**
   * Print formatted string to character buffer.
   * This sprintf only accepts one argument to print.
   * See bprintf how to print multiple arguments.
   *
   * format specifiers for byte, word, long: %c,%d,%i,%b,%o,%u,%x
   * format specifier for string: %s
   * field specification (example)
   *   %-05.7d where - for left justify, default right justify
   *                 0 for padding with zeroes, default spaces
   *                 5 minimum field width
   *                 . field separator
   *                 7 maximum field width
   *
   * An original C sprintf statement like:
   * sprintf(buffer,"outside temperature %d celsius",24);
   * would be written as:
   * OBJ
   *   fmt: "Format"
   * VAR
   *   byte buffer[128]
   * PUB
   *   fmt.sprintf(@buffer,string("outside temperature %d celsius"),24)
   *
   * @param str Character array for formatted output
   * @param fmt String defining format
   * @param arg value parameter
   * @return Index in str pointing beyond last written character (eg. size of output string)
   */
  index = bprintf(str,0L,fmt,arg);
  *(char *)(str+index) = 0;
  return index;
}

