/*
********************************
* jvmNative - native functions *
********************************
*/
#include <propeller.h>
#include "jvmComm.h"
#include "jvmData.h"
#include "jvmStack.h"
#include "jvmVirtualPeripheral.h"
#include "jvmDefines.h"
#include "jvmNative.h"

// char eeByte;  //to support eeprom read/write ... defined in jvmData.c
     
static long propPin(long portPin) {
  long pin,port;
/* convert javelin portPin format to propeller pin format */
  port = portPin >> 8;
  pin = encode(portPin & 0xFF);
  switch (port) {
    case 1: pin += 8;
	        break;
    case 2: pin += 16;
	        break;
    case 3: pin += 24;
	        break;
	//default:
  }
  return pin;
}

static void nm_core_CPU_writeObject(long address, long value) {
/* Used by stamp\core\Uart.java method start() to write 256byte buffer object reference into nmUartBufferH/L */
/* So nmUartBufferH/L holds address of byte array object, buffer really starts at +4 offset */
  writeVPregister(address, value >> 8); //store highbyte
  writeVPregister(address+1, value);    //store lowbyte
}

static void nm_core_CPU_delay(long period) {
  long t;
  /**
   * Delays the program execution for a period of time. The delay method does not
   * return until the delay period has expired. During a delay any installed Virtual
   * Peripherals will continue to run.
   * <p>
   *
   * @param period the length of time to delay for, measured in
   * 95.48 microsecond units.
   */
  t = CNT;
#ifdef INTERNAL_ERROR
  while ((CNT - t) / (CLKFREQ * 95.48 / 1000000) <= period) ;
#endif
}

  
static long nm_core_CPU_count(long timeout, long portPin, long edge) {
  long value,pin,t;
  /**
   * Count the number of rising or falling edge transitions that occur during
   * a set period of time.
   * If the pin is an output then it will be changed to be an input and will remain as
   * an input when the counting is complete.
   * <p>
   * To ensure reliable detection the input signal must maintain a value for at least
   * 8.68us. This means that a square wave must have a period of at least 17.36us and
   * a frequency less than 57.6kHz.
   *
   * @param timeout the amount of time to count transitions over. Measured in
   *                8.68us units. The value of 32,767 allows a maximum measurement
   *                time of 284.4ms.
   * @param portPin     the I/O pin to measure transitions on.
   * @param edge    the type of edge to count. Pass true to count rising edges,
   *                or false to count falling edges.
   * @return the number of rising edges or falling edges that occured on the pin
   *         during the timeout period.
   */
  /* adapted from object BS2_functions PUB COUNT (Pin, Duration) : Value */
  /* only does rising-edge pulses now  */
  edge++; //to keep compiler happy
  pin = propPin(portPin);
  DIRA &= ~(1<<pin);                                 // Set as input
  CTRA = 0;                                          // Clear any value in ctrb
                                                     // set up counter, pos edge trigger
  CTRA = (0b01010 << 26 ) | (0b001 << 23) | (0 << 9) | (pin);  
  FRQA = 1;                                          //1 count/trigger
  PSHA = 0;                                          // Clear phase - holds accumulated count
  t = CNT;
#ifdef INTERNAL_ERROR
  while ((CNT - t) / (CLKFREQ * 8.68 / 1000000) <= timeout) ; // Allow to count for duration in 8.68us units 
#endif
  value = PSHA;                                      // Return total count
  CTRA = 0;                                          // clear counter mode
  return value;
}

static long nm_core_CPU_rcTime(long timeout, long portPin, long pinState) {
  long value,pin, ClkStart,clkStop,pinvar,state;
  /**
   * Measure the length of time it takes for a pin to match a desired value.
   * If the pin is an output then it will be changed to be an input and will
   * remain as an input when rcTime is complete. rcTime can be used to measure
   * the charge/discharge time of a RC (resistor-capacitor) circuit in order to
   * read an analogue value.
   *
   * @param timeout the maximum measurable time.
   * @param portPin the I/O pin to measure.
   * @param pinState the pin state to wait for.
   * @return the amount of time taken for the pin to change to the desired value,
   * measured in 8.68us units. A value of -1 indicates that the timeout expired
   * before the desired state was reached.
   */
  /* adapted from object BS2_functions PUB RCTIME (Pin,State):Duration | ClkStart, ClkStop */
  timeout++; //to keep compiler happy
  pin = propPin(portPin);
  if (pinState)
    pinState = 1;
  else
    pinState = 0;
  DIRA &= ~(1<<pin);                                     //set as input
  ClkStart = CNT;                                        // Save counter for start time
  pinvar = encode(pin);
  state = pinState << pin;
  WAITPNE(state,pinvar);              // Wait for opposite state to end
  clkStop = CNT;                                         // Save stop time
#ifdef INTERNAL_ERROR
  value = (clkStop - ClkStart)/(CLKFREQ * 8.68 /1000000); // calculate in 8.68us resolution
#endif
  return value;
}

static long nm_core_CPU_pulseIn(long timeout, long portPin, long pinState) {
  long value,pin,pinvar,state;
  /**
   * Measure the length of a single pulse. Waits for the value of the selected
   * pin to match the value selected by the state parameter and then times how
   * long it takes for the pin to return to its original state.
   * If the pin is an output then it will be changed to be an input and will
   * remain as an input when complete.
   * <p>
   * A pulse of less than 8.68us in duration may be missed.
   *
   * @param timeout  the maximum amount of time to spend in the pulseIn method.
   *                 The timeout covers the complete time of the process both
   *                 the wait for the start of the pulse and the time of the pulse.
   * @param portPin      the I/O pin to measure the pulse on.
   * @param pinState    whether to measure a high pulse (true) or a low pulse (false).
   * @return the length of the pulse in 8.68us units. A value of 0 indicates
   *         that no start edge was detected. A value of -1 indicates that no
   *         stop edge was detected.
   */
  /* adapted from object BS2_functions PUB PULSIN_Clk(Pin, State) : Duration */
  /* need to implement timeout */
  timeout++; //to keep compiler happy
  pin = propPin(portPin);
  DIRA &= ~(1<<pin);
  CTRA = 0;
  if (pinState)
    pinState = 1;
  else
    pinState = 0;
  if (pinState == 1)
    CTRA = (0b11010 << 26 ) | (0b001 << 23) | (0 << 9) | (pin); // set up counter, A level count
  else
    CTRA = (0b10101 << 26 ) | (0b001 << 23) | (0 << 9) | (pin); // set up counter, !A level count
  FRQA = 1;
  pinvar = encode(pin);
  state = pinState << pin;
  WAITPNE(state, pinvar);              // Wait for opposite state ready
  PSHA = 0;                                              // Clear count
  WAITPEQ(state, pinvar);              // wait for pulse
  WAITPNE(state, pinvar);              // Wait for pulse to end
#ifdef INTERNAL_ERROR
  value = PSHA * 1000000 /CLKFREQ / 8.68;                // Return duration as 8.68us units
#endif
  CTRA = 0;                                              // stop counter
  return value;
}

static void nm_core_CPU_pulseOut(long length, long portPin) {
  long ClkCycles,pin;
  /**
   * Generate a pulse of a specific duration. If the pin is an input then it will
   * be changed to be an output and remain as an output when complete. To
   * ensure that a positive going pulse or negative going pulse is generated it
   * is a good idea to write a value to the pin first to put it into the correct
   * state. pulseOut inverts the value of the selected pin, waits for the
   * specified length of time, and then returns the pin to its original value.
   * <p>
   * The time is measured in 8.68us units with a maximum value of 32767 giving a
   * maximum pulse time of 284.4ms. A time of 0 will not cause a pulse.
   * pulseOut can be used to generate the control pulses for a servo motor.
   *
   * @param length the length of the pulse to generate, measured in 8.68us units.
   * @param portPin    the I/O to generate the pulse on.
   */
  /* adapted from object BS2_functions PUB PULSOUT(Pin,Duration)  | clkcycles */
  pin = propPin(portPin);
#ifdef INTERNAL_ERROR
  ClkCycles = (length * CLKFREQ / 1000000 * 8.68 - 1250);
#endif
  if (ClkCycles < 400) ClkCycles = 400; // 400 = Minimum waitcnt value to prevent lock-up
                                        // duration * clk cycles for 8.68us
                                        // - inst. time, min cntMin
  DIRA |= (1<<pin);                     // Set to output                                         
  OUTA ^= (1<<pin);                     // set to opposite state
  wait(ClkCycles + CNT);                // wait until clk gets there 
  OUTA ^= (1<<pin);                     // return to orig. state
}

static long nm_core_CPU_readPin(long portPin) {
  long pin;
  /**
   * Reads the logic value on a pin. If the pin is an output then it will be
   * changed to be an input and will remain as an input when complete.
   *
   * @param portPin the I/O pin to read.
   * @return the state of the pin where true indicates high and false indicates low.
   */
  pin = propPin(portPin);
  DIRA &= ~(1<<pin); //set as input
  return INA & (1<<pin);
}  

static void nm_core_CPU_writePin(long portPin, long value) {
  long pin;
  /**
   * Set a pin to a logic state. Makes the selected pin high or low as selected
   * by the state parameter. If the pin is an input then it will be changed to
   * be an output and remain as an output when complete.
   *
   * @param portPin the I/O pin to write.
   * @param value the state to set the pin to (true=high, false=low).
   */
  pin = propPin(portPin);
  DIRA |= (1<<pin);
  if (value & 1)
    OUTA |= (1<<pin);
  else
	OUTA &= ~(1<<pin);
}

//shiftin modes
#define  PRE_CLOCK_MSB   0  //Specify pre-clock sampling and MSB-first transmission. 
#define  POST_CLOCK_MSB  1  //Specify post-clock sampling and MSB-first transmission.
#define  PRE_CLOCK_LSB   2  //Specify pre-clock sampling and LSB-first transmission.
#define  POST_CLOCK_LSB  3  //Specify post-clock sampling and LSB-first transmission.
#define  OnClock         4

static long nm_core_CPU_shiftIn(long dataPortPin, long clockPortPin, long Bitcount, long Mode) {
  long value,Dpin,Cpin,InBit;
  /**
   * Read data from a synchronous serial device.
   * The clock pin outputs a clock signal and the data is clocked in form the data pin.
   * If the data pin is an output then it will be changed to be an input and remain as
   * an input when complete. If the clock pin is an input then it will be changed to be
   * an output and remain as an output when complete.
   * <p> The mode parameter selects pre/post clock sampling. Note that some devices
   * require pre-clock sampling. When pre-clock sampling is used the data pin is
   * sampled to obtain the first bit prior to any change of the clock pin. Clearly
   * the device must have had a clock pulse to output the first bit and this is
   * normally supplied by the previous call to shiftIn or shiftOut. Most
   * synchronous serial devices require a write (shiftOut) prior to reading and so the
   * clock for the first bit will be supplied by the shiftOut call. If the device must be
   * read without first being written to then the first read (shiftIn) should request an
   * extra bit to provide the initial clock pulse.
   * <p>
   * Note that the bit count can be set to higher than 16 (up to 256) if desired. When
   * a number greater than 16 is used the last 16 bits will be returned.
   *
   * @param dataPortPin the I/O pin to read data on.
   * @param clockPortPin the I/O pin on which to provide the clock to the external device.
   * @param bitCount the number of bits to count in (from 1 to 16). bitCount must not be 0.
   * @param mode the shifting and clocking mode to employ. This determines whether
   * bits are shifted into the least-significant-bit (LSb) or most-significant-bit (MSb)
   * first. The clocking mode determines whether the data is sampled before or after
   * the first clock transition. See the description for the possible values.
   * @return the bits shifted in on the data pin. The result contains bitCount bits and is
   * justified according to the mode setting. If MSb first is selected then the result
   * will be the data read in on the port. If LSb first is selected then the data is
   * returned MSb justified and must be shifted right (16-bitCount) bits to
   * obtain the correct value.
   */
  /* adapted from object BS2_functions PUB SHIFTIN (Dpin, Cpin, Mode, Bits) : Value | InBit */
/*
   Shift data in, master clock, for mode use BS2#MSBPRE, #MSBPOST, #LSBPRE, #LSBPOST
   Clock rate is ~16Kbps.  Use at 80MHz only is recommended.
     X = BS2.SHIFTIN(5,6,BS2#MSBPOST,8)
*/
  Dpin = propPin(dataPortPin);
  Cpin = propPin(clockPortPin);
  DIRA &= ~(1<<Dpin);                                    // Set data pin to input
  OUTA &= ~(1<<Cpin);                                   // Set clock low 
  DIRA |= (1<<Cpin);                                    // Set clock pin to output 
  if (Mode == PRE_CLOCK_MSB) {                            // Mode - MSB, before clock
    value = 0;
    while (Bitcount-- > 0) {                              // for number of bits
      InBit = (INA & (1<<Dpin)) != 0;                    // get bit value
      value = (value << 1) + (InBit & 1);                 // Add to  value shifted by position
      OUTA ^= (1<<Cpin);                                  // cycle clock
      OUTA ^= (1<<Cpin);
      wait(1000 + CNT);                                   // time delay
    }
  }
  else if (Mode == POST_CLOCK_MSB) {                      // Mode - MSB, after clock              
    value = 0;                                                          
    while (Bitcount-- > 0) {                              // for number of bits                    
      OUTA ^= (1<<Cpin);                                  // cycle clock                         
      OUTA ^= (1<<Cpin);                                         
      InBit = (INA & (1<<Dpin)) != 0;                     // get bit value                          
      value = (value << 1) + (InBit & 1);                 // Add to  value shifted by position                                         
      wait(1000 + CNT);                                   // time delay                            
    }
  }                                                               
  else if (Mode == POST_CLOCK_LSB) {                      // Mode - LSB, after clock                    
    value = 0;                                                                                         
    while (Bitcount-- > 0) {                              // for number of bits                         
      OUTA ^= (1<<Cpin);                                  // cycle clock                          
      OUTA ^= (1<<Cpin);                                                                             
      InBit = (INA & (1<<Dpin)) != 0;                     // get bit value                        
      value = ((InBit & 1) << (Bitcount-1)) + (value >> 1); // Add to  value shifted by position    
      wait(1000 + CNT);                                   // time delay 
	}                          
  }
  else if (Mode == PRE_CLOCK_LSB) {                       // Mode - LSB, before clock             
    value = 0;                                                                                   
    while (Bitcount-- > 0) {                              // for number of bits                   
      InBit = (INA & (1<<Dpin)) != 0;                     // get bit value                        
      value = (value >> 1) + ((InBit & 1) << (Bitcount-1)); // Add to  value shifted by position    
      OUTA ^= (1<<Cpin);                                  // cycle clock                          
      OUTA ^= (1<<Cpin);                                                                             
      wait(1000 + CNT);                                   // time delay
    }
  }
  else if (Mode == OnClock) {  /* is this required ?? */                                          
    value = 0;
    while (Bitcount-- > 0) {                              // for number of bits
      OUTA ^= (1<<Cpin);                                  // cycle clock
      wait(500 + CNT);                                    // get bit value
      InBit = (INA & (1<<Dpin)) != 0;                     // time delay
      value = (value << 1) + (InBit & 1);                 // Add to  value shifted by position
      OUTA ^= (1<<Cpin);
      wait(500 + CNT);                           
    } 
  }
  return value;
}

//shiftout modes
#define  SHIFT_MSB  0  //Specify MSB-first transmission.
#define  SHIFT_LSB  2  //Specify LSB-first transmission. 

static void nm_core_CPU_shiftOut(long dataPortPin, long clockPortPin, long Bitcount, long Mode, long data) {
  long Dpin,Cpin;
  /**
   * Write data to a synchronous serial device.
   * The clock pin outputs a clock signal and the data is clocked out the data pin. If
   * the data pin is an input then it will be changed to be an output and remain as an
   * output when complete. Likewise if the clock pin is an input then it will be
   * changed to be an output and remain as an output when complete.
   * <p>
   * Both the clock and data pins should be initialized before calling shiftOut. shiftOut
   * works by inverting the pins so you can generate either idle high or idle low signals.
   * <p>
   * If LSb first is selected then the data can be passed directly. If MSb first is
   * selected then if less than 16 bits are to be shift out, data should be shifted left
   * to MSb justify it.
   * @param dataPortPin the I/O pin to write data on.
   * @param clockPortPin the I/O pin on which to provide the clock to the external device.
   * @param bitCount the number of bits to count out (from 1 to 16). bitCount must not be 0.
   * @param mode the shifting mode to employ. This determines whether
   * bits are shifted into the least-significant-bit (LSb) or most-significant-bit (MSb)
   * first. Valid values are CPU.SHIFT_MSB and CPU.SHIFT_LSB.
   */
  /* adapted from object BS2_functions PUB SHIFTOUT (Dpin, Cpin, Value, Mode, Bits)| bitNum  */
/*
   Shift data out, master clock, for mode use ObjName#LSBFIRST, #MSBFIRST
   Clock rate is ~16Kbps.  Use at 80MHz only is recommended.
     BS2.SHIFTOUT(5,6,"B",BS2#LSBFIRST,8)
*/
  Dpin = propPin(dataPortPin);
  Cpin = propPin(clockPortPin);
  OUTA &= ~(1<<Dpin);                                   // Data pin = 0
  DIRA |= (1<<Dpin);                                     // Set data as output
  OUTA &= ~(1<<Cpin);
  DIRA |= (1<<Cpin);

  if (Mode == SHIFT_LSB) {                               // Send LSB first    
    while (Bitcount-- > 0) {
      if (data & 1)
	    OUTA |= (1<<Dpin);                              // Set output
	  else
	    OUTA &= ~(1<<Dpin);
      data = data >> 1;                                 // Shift value right
      OUTA ^= (1<<Cpin);                                // cycle clock
      OUTA ^= (1<<Cpin);
      wait(1000 + CNT);                                 // delay
    }
  }
  else if (Mode == SHIFT_MSB) {                         // Send MSB first               
    while (Bitcount-- > 0) {                                                                
      if (data & 0x80)
	    OUTA |= (1<<Dpin);                              // Set output
	  else
	    OUTA &= ~(1<<Dpin);           
      data = data << 1;                                 // Shift value left    
      OUTA ^= (1<<Cpin);                                // cycle clock          
      OUTA ^= (1<<Cpin);                                                             
      wait(1000 + CNT);                                 // delay                
    }
	OUTA &= ~(1<<Dpin);                                 // Set data to low
  }   
}
  
static void nm_core_CPU_uninstall(long bank) {
  /**
   * This function is called by CPU.removeVP. 
   * Remove a virtual peripheral from its slot. The VP will stop executing
   * immediately and its slot will be available to install a different VP.
   *
   * @param bank the rambank allocated to the virtual peripheral that must stop.
   */
  writeVPregister(bank + nmPeriphISRVec, 0);
}

static void nm_core_CPU_install(long routine) {
  //Appears not to be used
}
  
static void nm_core_CPU_nap(long period) {
  long t,delayms;
  //If the Prop cannot support nap, then just do a delay
  /**
   * Put the Prop into sleep mode for a period of time. The accuracy
   * of the sleep timer is very poor. See the manual for more information.
   *
   * @param period specifies the period for which the Prop should sleep. The
   * following values are valid:
   *
   * 0 - 16ms
   * 1 - 32ms
   * 2 - 64ms
   * 3 - 128ms
   * 4 - 256ms
   * 5 - 512ms
   * 6 - 1024ms
   * 7 - 2048ms
   */
  delayms = 16 << period; //number of milliseconds to delay
  t = CNT;
  while ((CNT - t) / (CLKFREQ/1000) < delayms) ;
}

static void nm_core_CPU_setOutput(long portPin) {
  /**
   * Make a pin an output.
   * The specified pin will be converted to an output. More than one pin can
   * be specified using the | operator, as long as all pins are on the same
   * port.
   *
   * @param portPin the pin(s) to make into an output.
   */
  if (portPin < 256)
    DIRA |= portPin;
  else {
    portPin &= 0xFF;
    DIRA |= (portPin<<8);
  }
}
    

static void nm_core_CPU_setInput(long portPin) {
  /**
   * Make a pin an input.
   * The specified pin will be converted to an input. More than one pin can
   * be specified using the | operator, as long as all pins are on the same
   * port.
   *
   * @param portPin the pin(s) to make into an input.
   */
  if (portPin < 256)
    DIRA &= (255 ^ portPin) | 0xFFFFFF00;
  else {
    portPin &= 0xFF;
    DIRA &= ((255 ^ portPin)<<8) | 0xFFFF00FF;
  }
}

static long nm_core_CPU_readSRAM(long address) {
  long value;
  /**
   * Reads a byte from SRAM at the specified address.
   * The address will be mapped into the device memory range.
   * This is an internal function and should not be used.
   *
   * @param address the address to read the byte from.
   * @return the byte read from the SRAM.
   */
  address %= getMemSize(); //to compensate for size 0x8000 in true javelin
  return readJavaByte(address);
}  

static void nm_core_CPU_writeSRAM(long address, long value) {
  /**
   * Writes a byte into the SRAM at the specified address.
   * The address will be mapped into the device memory range.
   * This is an internal function and should not be used.
   *
   * @param address the address to write the byte to.
   * @param value value to write.
   */
  address %= getMemSize(); //to compensate for size 0x8000 in true javelin
  writeJavaByte(address,value);
}  

static long nm_core_EEPROM_eeRead(long address) {
  long value,ack,temp;
  /**
   * Reads a byte from the EEPROM at the specified address.
   *
   * @param address the address to read the byte from.
   * @return the byte read from the EEPROM.
   */
  address -= (0x8000 - getMemSize()); //to compensate for constant 0x7FFF used in EEPROM.java
  return eeReadByte(address);
}        

static void nm_core_EEPROM_eeWrite(long address, long value) {
  long ack; 
  /**
   * Writes a byte into the EEPROM at the specified address.
   *
   * @param address the address to write the byte to.
   * @param value value to write.
   */
  address -= (0x8000 - getMemSize()); //to compensate for constant 0x7FFF used in EEPROM.java
  eeWriteByte(address,value);
}           

static void nm_core_Callback_returnTwo(void) {
  //This appears not to be used
}
  
static void nm_page_(void) {
  //Since the Prop has no pages, simply ignore ???
}

void doNative(long nm_function) {
  long arg0,arg1,arg2,arg3,arg4,retval;
/*
   Execute a native function. Arguments for native functions are either in VP registers
   or on the stack. If a native function returns a value it must be placed on top of the stack.
*/
  switch (nm_function) {
    case  0: arg1 = pop();
             arg0 = pop();
             if ((arg0>=nmPeriph0Bank) && (arg0<nmPeriph0Bank+MAX_NUM_VPS*16)) writeVPregister(arg0,arg1);
             else writeJVMregister(arg0,arg1);
			 break;
    case  1: arg0 = pop();
             if ((arg0>=nmPeriph0Bank) && (arg0<nmPeriph0Bank+MAX_NUM_VPS*16)) retval = readVPregister(arg0);
             else retval = readJVMregister(arg0);
             push(retval);
			 break;
    case  2: arg1 = pop();
             arg0 = pop();
             nm_core_CPU_writeObject(arg0,arg1);
			 break;
    case  3: arg0 = pop();
             nm_core_CPU_delay(arg0);
			 break;
    case  4: arg1 = pop();
             arg0 = pop();
             message(arg0,arg1);
			 break;
    case  5: nm_page_();
	         break;
    case  6: arg2 = pop();
             arg1 = pop();
             arg0 = pop();
             nm_core_CPU_count(arg0,arg1,arg2);
			 break;
    case  7: arg2 = pop();
             arg1 = pop();
             arg0 = pop();
             retval = nm_core_CPU_rcTime(arg0,arg1,arg2);
			 push(retval);
			 break;
    case  8: arg2 = pop();
             arg1 = pop();
             arg0 = pop();
             retval = nm_core_CPU_pulseIn(arg0,arg1,arg2);
             push(retval);
			 break;
    case  9: arg1 = pop();
             arg0 = pop();
             nm_core_CPU_pulseOut(arg0,arg1);
			 break;
    case 10: arg0 = pop();
             retval = nm_core_CPU_readPin(arg0);
             push(retval);
			 break;
    case 11: arg1 = pop();
             arg0 = pop();
             nm_core_CPU_writePin(arg0,arg1);
			 break;
    case 12: arg3 = pop();
             arg2 = pop();
             arg1 = pop();
             arg0 = pop();
             retval = nm_core_CPU_shiftIn(arg0,arg1,arg2,arg3);
             push(retval);
			 break;
    case 13: arg4 = pop();
             arg3 = pop();
             arg2 = pop();
             arg1 = pop();
             arg0 = pop();
             nm_core_CPU_shiftOut(arg0,arg1,arg2,arg3,arg4);
			 break;
    case 14: retval = isCarry();
             push(retval);
			 break;
    case 15: nm_page_();
			 break;
    case 16: arg0 = pop();
             nm_core_CPU_uninstall(arg0);
			 break;
    case 17: arg0 = pop();
             nm_core_CPU_install(arg0);
			 break;
    case 18: arg0 = pop();
             nm_core_CPU_nap(arg0);
			 break;
    case 19: arg0 = pop();
             nm_core_CPU_setOutput(arg0);
			 break;
    case 20: nm_page_();
			 break;
    case 21: arg0 = pop();
             nm_core_CPU_setInput(arg0);
			 break;
    case 22: nm_page_();
			 break;
    case 23: arg0 = pop();
             retval = nm_core_CPU_readSRAM(arg0);
             push(retval);
			 break;
    case 24: nm_page_();
			 break;
    case 25: arg1 = pop();
             arg0 = pop();
             nm_core_CPU_writeSRAM(arg0,arg1);
			 break;
    case 26: nm_page_();
			 break;
    case 27: arg0 = pop();
             retval = nm_core_EEPROM_eeRead(arg0);
             push(retval);
			 break;
    case 28: nm_page_();
			 break;
    case 29: arg1 = pop();
             arg0 = pop();
             nm_core_EEPROM_eeWrite(arg0,arg1);
			 break;
    case 30: nm_page_();
			 break;
    case 31: arg0 = pop();
             nm_core_Uart_txInit(arg0);
			 break;
    case 32: arg0 = pop();
             nm_core_Uart_rxInit(arg0);
			 break;
    case 33: arg0 = pop();
             nm_core_Uart_rxRead(arg0);
			 break;
    case 34: nm_core_Callback_returnTwo();
			 break;
    case 35: arg0 = pop();
             nm_core_Timer_init(arg0);
			 break;
    case 36: arg0 = pop();
             nm_core_Timer_latch(arg0);
			 break;
    case 37: nm_page_();
			 break;
    case 38: arg0 = pop();
             nm_core_PWM_install(arg0);
			 break;
    case 39: nm_page_();
			 break;
    case 40: arg4 = pop();
             arg3 = pop();
             arg2 = pop();
             arg1 = pop();
             arg0 = pop();
             nm_core_PWM_updateInternal(arg0,arg1,arg2,arg3,arg4);
			 break;
    case 41: nm_page_();
			 break;
    case 42: arg0 = pop();
             nm_core_DAC_install(arg0);
			 break;
    case 43: nm_page_();
			 break;
    case 44: arg0 = pop();
             nm_core_ADC_install(arg0);
			 break;
    case 45: nm_page_();
			 break;
    case 46: retval = getByte();
             push(retval);
			 break;
    case 47: nm_page_();
			 break;
    //default:
  }
}


  