/*
**********************************************************
* File jvmVirtualPeripheral: native routines for VP code *
**********************************************************
*/

/*
  Virtual Peripherals are implemented as background code, meaning the code executes
  independantly from the interpreted java code.
*/
#include <propeller.h>
#include "jvmDefines.h"
#include "jvmVpCore.h"
//#include "jvmVpCore.h"
#include "jvmVpQueue.h"
#include "jvmVirtualPeripheral.h"  
  
//do not change order
long   vpTimerCount;                //holds 32bit free running timer
long   vpRambank[MAX_NUM_VPS*6];    //6 longs per VP slot, first 16 bytes are VP registers, last 2 longs are input pinmask and output pinmask
long   vpQueueStack[20];
long   vpQid;
      
void vpStart(void) {
/* startup COG's for VP code */
  vpTimerCount = firstBankVPcog1;  //first bank of registers for vpCog1 is at 54 bytes offset to vpTimerCount
  //vpCog1.start(@vpTimerCount)
  vpTimerCount = firstBankVPcog2;  //first bank of registers for vpCog2 is at 102 bytes offset to vpTimerCount
  //vpCog2.start(@vpTimerCount)
  //vpQid = cognew(qStart(vpRambank),vpQueueStack)+1; //run queue code in seperate cog
}  
  
void vpStop(void) {
/* stop COG's for VP code */
  COGSTOP(vpQid-1);
  //vpCog2.stop
  //vpCog1.stop
}  
  
void writeVPregister(long address, long value) {
  long bank,offset;
/* write value to vpRambank register */
  bank = (address & 0xF0);
  offset = (address & 0x0F);
  *(char *)(vpRambank + (((bank - nmPeriph0Bank)*3)/2) + offset) = value;
}  
  
long readVPregister(long address) {
  long bank,offset;
/* read value from vpRambank register */
  bank = (address & 0xF0);
  offset = (address & 0x0F);
  return *(char *)(vpRambank + (((bank - nmPeriph0Bank)*3)/2) + offset) & 0xFF;
}

void nm_core_Timer_init(long bank) {
  /**
   * Initialize the 32bit counter. The 32bit nmTimer register should increment every 8.68 usec.
   * Called by Timer.start()
   *
   * @param bank The allocated area for the Timer registers.
   */
  vpTimerCount = 0;
  writeVPregister(bank + nmPeriphISRVec, TIMER);
}  

void nm_core_Timer_latch(long bank) {
  long latchedCount;
  /**
   * Latch the 32bit counter into the latch registers.
   * Called by Timer.tickLo()
   *
   * @param bank The allocated area for the Timer registers.
   */
  latchedCount = vpTimerCount;                               //get current count 
  writeVPregister(bank + nmLatchTimer1, latchedCount      ); //write b0
  writeVPregister(bank + nmLatchTimer2, latchedCount >> 8 ); //write b1
  writeVPregister(bank + nmLatchTimer3, latchedCount >> 16); //write b2
  writeVPregister(bank + nmLatchTimer4, latchedCount >> 24); //write b3
}  

void nm_core_ADC_install(long bank) {
  long slot,pin,port,pinmask;
  /**
   * Initialize an ADC as background task. Its parameters are in vpRambank.
   * Called by ADC.start()
   *
   * @param bank The allocated area for the ADC registers.
   */
  slot = (bank - nmPeriph0Bank) >> 4; //bank 0x80..0xD0 converted to slot 0..5
  pin = readVPregister(bank + nmInPin);
  port = readVPregister(bank + nmInPort);
  if (pin == 0)
    pinmask = 0;
  else
    pinmask = pin << (port<<3);
  vpRambank[slot*6 + 4] = pinmask;
  pin = readVPregister(bank + nmOutPin);
  port = readVPregister(bank + nmOutPort);
  if (pin == 0)
    pinmask = 0;
  else
    pinmask = pin << (port<<3);
  vpRambank[slot*6 + 5] = pinmask;
  writeVPregister(bank + nmPeriphISRVec, ADC); //start adc VP
}
  
void nm_core_DAC_install(long bank) {
  long slot,pin,port,pinmask;
  /**
   * Initialize a DAC as background task. Its parameters are in vpRambank.
   * Called by DAC.start()
   *
   * @param bank The allocated area for the DAC registers.
   */
  slot = (bank - nmPeriph0Bank) >> 4; //bank 0x80..0xD0 converted to slot 0..5
  vpRambank[slot*6 + 4] = 0; //no input pin
  pin = readVPregister(bank + nmDacPin);
  port = readVPregister(bank + nmDacPort);
  if (pin == 0)
    pinmask = 0;
  else
    pinmask = pin << (port<<3);
  vpRambank[slot*6 + 5] = pinmask;
  writeVPregister(bank + nmDacAccu, 0);
  writeVPregister(bank + nmPeriphISRVec, DAC); //start dac VP
}

void nm_core_PWM_install(long bank) {
  long slot,pin,port,pinmask;
  /**
   * Initialize a PWM as background task. Its parameters are in vpRambank.
   * Called by PWM.start()
   *
   * @param bank The allocated area for the PWM registers.
   */
  slot = (bank - nmPeriph0Bank) >> 4; //bank 0x80..0xD0 converted to slot 0..5
  vpRambank[slot*6 + 4] = 0; //no input pin
  pin = readVPregister(bank + nmPWMPin);
  port = readVPregister(bank + nmPWMPort);
  if (pin == 0)
    pinmask = 0;
  else
    pinmask = pin << (port<<3);
  vpRambank[slot*6 + 5] = pinmask;
  writeVPregister(bank + nmPeriphISRVec, PWM); //start pwm VP
}

void nm_core_PWM_updateInternal(long bank, long highTime, long lowTime, long highZero, long lowZero) {
  /**
   * Update the PWM duty cycle. Its parameters are in vpRambank. 
   * Called by PWM.update()
   *
   * @param bank The allocated area for the PWM registers.
   * @param highTime the number of 8.68us the waveform should be high. (1-255)
   * @param lowTime the number of 8.68us the waveform should be low. (1-255)
   * @param highZero Boolean value (always true) that tells wether highTime != 0.
   * @param lowZero Boolean value (always true) that tells wether lowTime != 0.
   */
  writeVPregister(bank + nmPWMlowTime, lowTime);
  writeVPregister(bank + nmPWMhighTime, highTime);
  writeVPregister(bank + nmPWMlowZero, lowZero);
  writeVPregister(bank + nmPWMhighZero, highZero);
}       

void nm_core_Uart_rxInit(long bank) {
  long bufptr,slot,pin,port,pinmask,invert,stopBits;
  /**
   * Initialize a Uart receiver as background task. Its parameters are in vpRambank.
   * Called by Uart.start()
   *
   * @param bank The allocated area for the Uart registers.
   */
  slot = (bank - nmPeriph0Bank) >> 4; //bank 0x80..0xD0 converted to slot 0..5
  pin = readVPregister(bank + nmUartPin);
  port = readVPregister(bank + nmUartPort);
  if (pin == 0)
    pinmask = 0;
  else
    pinmask = pin << (port<<3);
  vpRambank[slot*6 + 4] = pinmask;
  pin = readVPregister(bank + nmUartHSPin);
  port = readVPregister(bank + nmUartHSPort);
  if (pin == 0)
    pinmask = 0;
  else
    pinmask = pin << (port<<3);
  vpRambank[slot*6 + 5] = pinmask;

  bufptr = (readVPregister(bank + nmUartBufferH)<<8) + readVPregister(bank + nmUartBufferL); //get buffer object reference
  bufptr += 4; //we do not need to check for null pointer, as a NullPointerException would be thrown before we get here
  writeVPregister(bank + nmUartBufferH, bufptr);   //store buffer address as little endian
  writeVPregister(bank + nmUartBufferL, bufptr>>8);

  stopBits = encode(readVPregister(bank + nmUartStopBits)); //nmUartStopBits is 1,3,7,15...255 so stopBits is 0 to 7
  stopBits += 10;  //make it 1 to 8 stopbits + startbit + 8 databits
  writeVPregister(bank + nmUartStopBits, stopBits); //store stopBits (=total bits) as decimal value
  writeVPregister(bank + nmPeriphISRVec, UARTRX);
}

void nm_core_Uart_rxRead(long bank) {
  long slot,pinmask,head,tail,len,invert;
  /**
   * Inform Uart receiver a byte has been grabbed from the buffer, so the receiver can do the handshaking.
   * Called by Uart.readByte()
   *
   * @param bank The allocated area for the Uart registers.
   */
  slot = (bank - nmPeriph0Bank) >> 4; //bank 0x80..0xD0 converted to slot 0..5
  pinmask = vpRambank[slot*6 + 5]; //get HS pinmask
  if (pinmask != 0) {
    invert = readVPregister(bank + nmUartInvert) & 0x02;
    head = readVPregister(bank + nmUartHead);
    tail = readVPregister(bank + nmUartTail);
    len = (head - tail + 256) % 256;
    if (len <= 16) {
      //enable hsPortpin (tell remote sender to resume transmitting)
      if (invert)
        OUTA |= pinmask;
      else
        OUTA &= ~pinmask;
    }
	if (len >= 240) {
      //disable hsPortPin (tell remote sender to stop transmitting)
      if (invert)
        OUTA &= ~pinmask;
      else
        OUTA |= pinmask;
    }
	DIRA |= pinmask; //make pin output
  } 
}
      
void nm_core_Uart_txInit(long bank) {
  long bufptr,slot,pin,port,pinmask,invert,stopBits;
  /**
   * Initialize a Uart transmitter as background task. Its parameters are in vpRambank.
   * Called by Uart.start()
   *
   * @param bank The allocated area for the Uart registers.
   */
  slot = (bank - nmPeriph0Bank) >> 4; //bank 0x80..0xD0 converted to slot 0..5
  pin = readVPregister(bank + nmUartHSPin);
  port = readVPregister(bank + nmUartHSPort);
  if (pin == 0)
    pinmask = 0;
  else
    pinmask = pin << (port<<3);
  vpRambank[slot*6 + 4] = pinmask;
  pin = readVPregister(bank + nmUartPin);
  port = readVPregister(bank + nmUartPort);
  if (pin == 0)
    pinmask = 0;
  else
    pinmask = pin << (port<<3);
  vpRambank[slot*6 + 5] = pinmask;

  bufptr = (readVPregister(bank + nmUartBufferH)<<8) + readVPregister(bank + nmUartBufferL); //get buffer object reference
  bufptr += 4; //we do not need to check for null pointer, as a NullPointerException would be thrown before we get here
  writeVPregister(bank + nmUartBufferH, bufptr);  //store buffer address as little endian
  writeVPregister(bank + nmUartBufferL, bufptr>>8);

  stopBits = encode(readVPregister(bank + nmUartStopBits)); //nmUartStopBits is 1,3,7,15...255 so stopBits is 0 to 7
  stopBits += 10; //make it 1 to 8 stopbits + startbit + 8 databits
  writeVPregister(bank + nmUartStopBits, stopBits); //store stopBits (=total bits) as decimal value
  writeVPregister(bank + nmPeriphISRVec, UARTTX);
}

  